/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "main.h"

int initModuleEx(HWND hParentWnd, HINSTANCE hInstance, LPCSTR szPath)
{
	hInst = hInstance;

	Window::init(hInst);

	// Bind Litestep logging function
	_LSLog = (BOOL (WINAPI*)(int,LPCSTR,LPCSTR))GetProcAddress(GetModuleHandle("LSAPI.DLL"), "_LSLog@12");
	if(!_LSLog) {
		// The export name might be fixed in the future
		_LSLog = (BOOL (WINAPI*)(int,LPCSTR,LPCSTR))GetProcAddress(GetModuleHandle("LSAPI.DLL"), "LSLog");
	}

	// Get the desktop window
	desktopWnd = FindWindow("DesktopBackgroundClass", NULL);
	// If the desktop window isn't found, wait a while and try again
	if(!desktopWnd) {
		Sleep(500);
		desktopWnd = FindWindow("DesktopBackgroundClass", NULL);
		if(!desktopWnd) {
			Log(LOG_ERROR, "Could not find desktop window");
			MessageBox(NULL, "Could not find desktop window", "IconDesk", MB_SETFOREGROUND);
			return 1;
		}
	}

	// Subclass the desktop
	oldWndProc = (WNDPROC)GetWindowLong(desktopWnd, GWL_WNDPROC);
	SetWindowLong(desktopWnd, GWL_WNDPROC, (LONG)wndProc);

	// Modify the desktop style to support double clicks
	oldClassStyle = GetClassLong(desktopWnd, GCL_STYLE);
	SetClassLong(desktopWnd, GCL_STYLE, GetClassLong(desktopWnd, GCL_STYLE) | CS_DBLCLKS);

	// Set up tooltop window
	tooltipWnd = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInstance, NULL);
	if(tooltipWnd) {
		SetWindowPos(tooltipWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	}
	tooltipCounter = 0;

	// Register bang commands
	AddBangCommand("!IconDeskShow", bangShow);
	AddBangCommand("!IconDeskHide", bangHide);
	AddBangCommand("!IconDeskToggle", bangToggle);
	AddBangCommand("!IconDeskUpdate", bangUpdate);

	// Read the config
	loadGroups();

	// Redraw the desktop
	InvalidateRect(desktopWnd, NULL, TRUE);

	return 0;
}

void quitModule(HINSTANCE hInstance)
{
	// Remove the bangs
	RemoveBangCommand("!IconDeskShow");
	RemoveBangCommand("!IconDeskHide");
	RemoveBangCommand("!IconDeskToggle");
	RemoveBangCommand("!IconDeskUpdate");

	// Delete the groups
	deleteGroups();

	// Get rid of the tooltips
	if(tooltipWnd) {
		DestroyWindow(tooltipWnd);
	}

	// Restore the desktop
	SetWindowLong(desktopWnd, GWL_WNDPROC, (LONG)oldWndProc);
	SetClassLong(desktopWnd, GCL_STYLE, oldClassStyle);

	// Redraw the desktop
	InvalidateRect(desktopWnd, NULL, TRUE);
}

LRESULT wndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg) {
		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
		case WM_MBUTTONUP:
		case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDBLCLK:
			{
				// Forward the message to all groups
				BOOL hit = FALSE;
				for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
					if((*it)->mouseEvent(uMsg, wParam, lParam)) {
						hit = TRUE;
					}
				}
				// Only forward the message if we didn't handle it
				if(hit) {
					return 0;
				}
			}
			break;

		case WM_ERASEBKGND:
			// Paint the desktop (if we let the desktop module do it, it'll flicker)
			PaintDesktop((HDC)wParam);

		case WM_PAINT:
			{
				PAINTSTRUCT	ps;
				HDC			hDC;

				if(!wParam) {
					hDC = BeginPaint(hWnd, &ps);
				}
				else {
					hDC = (HDC)wParam;
				}
				
				for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
					(*it)->paintIcons(hDC);
				}

				if(!wParam) {
					EndPaint(hWnd, &ps);
				}
			}
			if(uMsg == WM_ERASEBKGND) {
				return 1;
			}
			else if(uMsg == WM_PAINT) {
				return 0;
			}

	}
	return CallWindowProc(oldWndProc, hWnd, uMsg, wParam, lParam);
}

void bangShow(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->show();
	}
}

void bangHide(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->hide();
	}
}

void bangToggle(HWND caller, LPCSTR args)
{
	Group* group = getGroup(args);
	if(group) {
		group->toggle();
	}
}

void bangUpdate(HWND caller, LPCSTR args)
{
	if(strlen(args)) {
		Group* group = getGroup(args);
		if(group) {
			group->update();
		}
	}
	else {
		// If no group was specified, update all groups
		for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
			(*it)->update();
		}
	}
}

void loadGroups()
{
	char* group;
	int i = 0;

	FILE* fpStep = LCOpen(NULL);
	if(fpStep) {
		char szBuffer[MAX_LINE_LENGTH], szData[MAX_LINE_LENGTH], szDummy[MAX_LINE_LENGTH];
		char* rszTokens[1];

		rszTokens[0] = szDummy;

		szBuffer[0] = 0;
		while(LCReadNextConfig(fpStep, "IconDeskGroups", szBuffer, sizeof(szBuffer))) {
			szDummy[0] = szData[0] = 0;
			LCTokenize(szBuffer, rszTokens, 1, szData);

			group = strtok(szData, " ,\t");
			while(group) {
				groups.push_back(new Group(group, desktopWnd));
				group = strtok(NULL, " ,\t");
				i++;
			}
		}
		LCClose(fpStep);
	}
	
	if(i == 0) {
		groups.push_back(new Group("Default", desktopWnd));
	}
}

void deleteGroups()
{
	for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
		delete *it;
	}
	groups.clear();
}

Group* getGroup(const char* name)
{
	for(GroupListIter it = groups.begin(); it != groups.end(); it++) {
		if(!strcmpi((*it)->name, name)) {
			return *it;
		}
	}
	return NULL;
}

UINT AddTooltip(LPSTR szText, RECT rect, UINT id)
{
	TOOLINFO ti;

	if(!tooltipWnd) {
		return 0;
	}

	if(id == 0) {
		id = ++tooltipCounter;
	}

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = desktopWnd;
	ti.uId = id;
	ti.rect = rect;
	ti.hinst = hInst;
	ti.lpszText = szText;
	ti.lParam = 0;

	SendMessage(tooltipWnd, TTM_ADDTOOL, 0, LPARAM(&ti));

	return id;
}

void DeleteTooltip(UINT id)
{
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	if(!tooltipWnd) {
		return;
	}

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = desktopWnd;
	ti.uId = id;
	ti.rect = emptyRect;
	ti.hinst = hInst;
	ti.lpszText = NULL;
	ti.lParam = 0;

	SendMessage(tooltipWnd, TTM_DELTOOL, 0, LPARAM(&ti));
}

BOOL Log(int nLevel, LPCSTR pszMessage)
{
	if(_LSLog) {
		return _LSLog(nLevel, "IconDesk", pszMessage);
	}
	else {
		if(nLevel == LOG_ERROR) {
			MessageBox(NULL, pszMessage, "IconDesk", MB_SETFOREGROUND);
		}
		return TRUE;
	}
}

BOOL LogPrintf(int nLevel, LPCSTR pszFormat, ...)
{
    char szMessage[MAX_LINE_LENGTH];
    va_list argList;
    
    va_start(argList, pszFormat);
    wvsprintf(szMessage, pszFormat, argList);
    va_end(argList);
    
    return Log(nLevel, szMessage);
}
